/*
 * mdtm_monitor.c
 *
 *  Created on: Sep 1, 2015
 *      Author: liangz
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <arpa/inet.h>
#include "mdtm.h"
#include "mdtm_csv.h"
#include "mdtm_rrd.h"
#include "mdtm_io_thread.h"
#include "mdtm_monitor.h"
#include "../globus_i_gridftp_server.h"

static mdtm_csv_file_t *        g_csvfile = 0;
static mdtm_csv_file_t *        g_csvfile_compl = 0;
static mdtm_csv_file_t *        g_csvfile_service = 0;
static mdtm_rrd_file_t          *g_rrdfile_in = 0, *g_rrdfile_out = 0;
static void                     *g_rrddata_in = 0, *g_rrddata_out = 0;

/*
 *      Internal APIs
 */
static
void
mdtm_monitor_log_servicestatus()
{
//  struct mdtm_device_s *        devices;
//  int                           nDevices;
  mdtm_thread_desc_t *          thread_descs;
  int                           i, count = 0;
  static void *                 data = 0;
  static int                    done = 0;

  if(done || !g_csvfile_service) {
    return;
  }

//  if((nDevices = mdtm_config_getpcidevices(&devices)) <= 0)
//    return;

  count = mdtm_io_thread_get_threaddescs(&thread_descs);

  // check if the thread_descs are ready
  for(i = 0; i < count; i++) {
      if(thread_descs[i].numaindex < 0)
          return;
  }

  for(i = 0; i < count; i++) {
      data = mdtm_csvdata_alloc(g_csvfile_service);
      if(data == NULL)
        break;
      mdtm_csvfile_add_data(g_csvfile_service, data);
      mdtm_csvdata_set(data, "ThreadID", thread_descs[i].id.tid);
      mdtm_csvdata_set(data, "Type", thread_descs[i].type);
      mdtm_csvdata_set(data, "Numa", thread_descs[i].numaindex);
      mdtm_csvdata_set(data, "Core", thread_descs[i].boundcpu);
      mdtm_csvdata_set_str(data, "Device", thread_descs[i].devices[0]);
  }
  done = 1;
}

static
void
mdtm_monitor_log_update(mdtm_xmit_progress_t *xmit_progress)
{
  void *        data;
  int           current_time;
  size_t        rate;
  mdtm_cvs_aggr_t       aggr_in, aggr_out;

  // get the current Unix epoch
  current_time = time(0);

  // snapshot csv
  xmit_progress->file.csv_snapshot = g_csvfile;
  data = mdtm_csvfile_lookup_data(g_csvfile,
      "sid",
      xmit_progress->sid);
  if(!data) {
      // if historic session, just ignore
      if(g_csvfile_compl &&
          mdtm_csvfile_lookup_data(
              g_csvfile_compl,"sid",xmit_progress->sid))
        return;

      // new session, create and add data to snapshots
      data = mdtm_csvdata_alloc(g_csvfile);
      mdtm_csvfile_add_data(g_csvfile, data);
      rate = 0;
      mdtm_csvdata_set(data, "t0", current_time);
  }
  else {
      int delta_time = current_time - mdtm_csvdata_get(data, "ts");

      if(delta_time < 1)
        return;

      rate = (xmit_progress->xfr - mdtm_csvdata_get(data, "xfr")) / delta_time;
//      mdtm_csvdata_set_str(data, "rrd_loc", xmit_progress->filepath);
  }

  mdtm_csvdata_set(data, "ts", current_time);
  mdtm_csvdata_set(data, "sid", xmit_progress->sid);
  mdtm_csvdata_set(data, "uid", xmit_progress->uid);
//  mdtm_csvdata_set(data, "t0", current_time);
  mdtm_csvdata_set(data, "t1", current_time);
  mdtm_csvdata_set(data, "src", xmit_progress->src);
  mdtm_csvdata_set(data, "dst", xmit_progress->dst);
  mdtm_csvdata_set(data, "size", (long long)xmit_progress->size);
  mdtm_csvdata_set(data, "xfr", xmit_progress->xfr);
  mdtm_csvdata_set(data, "rate", rate);
  mdtm_csvdata_set(data, "dir", xmit_progress->in0out1);
  mdtm_csvdata_set_str(data, "rrd_loc", xmit_progress->filepath);
//  mdtm_csvfile_flush_data(xmit_progress->file.csv_snapshot);
  //    mdtm_csvfile_close(csvfile);

  // aggregate in and out rrd files
  mdtm_csvfile_aggregate_data(g_csvfile, &aggr_in, &aggr_out);
  xmit_progress->file.rrd_aggr_in = g_rrdfile_in;
  xmit_progress->file.rrd_aggr_out = g_rrdfile_out;
  mdtm_rrddata_set(g_rrddata_in, "nsid", aggr_in.nsid);
  mdtm_rrddata_set(g_rrddata_in, "size", aggr_in.size);
  mdtm_rrddata_set(g_rrddata_in, "xfr", aggr_in.xfr);
  mdtm_rrddata_set(g_rrddata_in, "rate", aggr_in.rate);
  mdtm_rrddata_set(g_rrddata_out, "nsid", aggr_out.nsid);
  mdtm_rrddata_set(g_rrddata_out, "size", aggr_out.size);
  mdtm_rrddata_set(g_rrddata_out, "xfr", aggr_out.xfr);
  mdtm_rrddata_set(g_rrddata_out, "rate", aggr_out.rate);
//  mdtm_rrdfile_write(xmit_progress->file.rrd_aggr, g_rrddata);

  // service status
  mdtm_monitor_log_servicestatus();
}

static
void
mdtm_monitor_log_finish(mdtm_xmit_progress_t *xmit_progress)
{
  void *        data;
  int           current_time;
  size_t        rate;
  mdtm_cvs_aggr_t       aggr_in, aggr_out;

  // get the current Unix epoch
  current_time = time(0);

  // snapshot csv
  xmit_progress->file.csv_snapshot = g_csvfile;
  data = mdtm_csvfile_lookup_data(g_csvfile,
      "sid",
      xmit_progress->sid);
  if(!data) {
//      data = mdtm_csvdata_alloc(g_csvfile);
//      mdtm_csvfile_add_data(g_csvfile, data);
//      rate = 0;
//      mdtm_csvdata_set(data, "t0", current_time);
      return;
  }
  else {
      int delta_time = current_time - mdtm_csvdata_get(data, "ts");

      if(delta_time < 1)
        return;

      rate = (xmit_progress->xfr - mdtm_csvdata_get(data, "xfr")) / delta_time;
//      mdtm_csvdata_set_str(data, "rrd_loc", xmit_progress->filepath);
  }

  mdtm_csvdata_set(data, "ts", current_time);
  mdtm_csvdata_set(data, "sid", xmit_progress->sid);
  mdtm_csvdata_set(data, "uid", xmit_progress->uid);
//  mdtm_csvdata_set(data, "t0", current_time);
  mdtm_csvdata_set(data, "t1", current_time);
  mdtm_csvdata_set(data, "src", xmit_progress->src);
  mdtm_csvdata_set(data, "dst", xmit_progress->dst);
  if(xmit_progress->in0out1) {
      mdtm_csvdata_set(data, "size", (long long)xmit_progress->size);
      mdtm_csvdata_set(data, "xfr", (long long)xmit_progress->size);
  }
  else{
      mdtm_csvdata_set(data, "size", (long long)xmit_progress->xfr);
      mdtm_csvdata_set(data, "xfr", (long long)xmit_progress->xfr);
  }
  mdtm_csvdata_set(data, "rate", rate);
  mdtm_csvdata_set(data, "dir", xmit_progress->in0out1);
  mdtm_csvdata_set_str(data, "rrd_loc", xmit_progress->filepath);

  // move the entry from snapshot to complete file.
  mdtm_csvfile_add_data(g_csvfile_compl, data);
  mdtm_csvfile_remove_data(g_csvfile, data);

  // save to files
  mdtm_csvfile_flush_data(g_csvfile_compl);
  mdtm_csvfile_flush_data(g_csvfile);

  // free the memory of completed data
  // TODO: LZ: we need to delete data older than a given time,
  //           i.e. 7 days, in the complete csv file
//  mdtm_csvfile_delete_data(g_csvfile_compl, timevalue);

//  mdtm_csvfile_close(file);
//  mdtm_csvfile_close(file1);

//  mdtm_csvfile_flush_data(xmit_progress->file.csv_snapshot);
  //    mdtm_csvfile_close(csvfile);

  // aggregate in and out rrd files
  mdtm_csvfile_aggregate_data(g_csvfile, &aggr_in, &aggr_out);

  xmit_progress->file.rrd_aggr_in = g_rrdfile_in;
  mdtm_rrddata_set(g_rrddata_in, "nsid", aggr_in.nsid);
  mdtm_rrddata_set(g_rrddata_in, "size", aggr_in.xfr);
  mdtm_rrddata_set(g_rrddata_in, "xfr", aggr_in.xfr);
  mdtm_rrddata_set(g_rrddata_in, "rate", aggr_in.rate);

  xmit_progress->file.rrd_aggr_out = g_rrdfile_out;
  mdtm_rrddata_set(g_rrddata_out, "nsid", aggr_out.nsid);
  mdtm_rrddata_set(g_rrddata_out, "size", aggr_out.size);
  mdtm_rrddata_set(g_rrddata_out, "xfr", aggr_out.size);
  mdtm_rrddata_set(g_rrddata_out, "rate", aggr_out.rate);
//  mdtm_rrdfile_write(xmit_progress->file.rrd_aggr, g_rrddata);

  // service status
  mdtm_monitor_log_servicestatus();
}

/*
 *      Public APIs
 */
char *
mdtm_monitor_getroot(int purge) {
  static char *         root = 0;
  char *                tmp = 0;
  int                   len;

  if(!purge && root && strcmp(root, ""))
    return strdup(root);

  tmp = globus_gfs_config_get_string("mdtm_monitor_path");

  if(!tmp || (len = strlen(tmp)) <= 0) {
      root = strdup(LOG_ROOT_DEFAULT);
  }
  else {
      if(tmp[len-1] != '/') {
          root = (char*) malloc(len + 2);
          strcpy(root, tmp);
          strcat(root, "/");
      }
      else
        root = strdup(tmp);
  }
  return root;
}

int
mdtm_monitor_getperiod() {
  return 10;
}

int
mdtm_monitor_enabled() {
  return globus_gfs_config_get_bool("mdtm_monitor") && mdtm_io_thrd_mgr_list_get("man");
}

void
mdtm_monitor_log_init()
{
  char        filepath[256], *fp, *root;

  root = mdtm_monitor_getroot(1);
  strcpy(filepath, root);
  if(root) free(root);
  fp = filepath + strlen(filepath);

  strcpy(fp, LOG_AGGR_RRD_IN);
  g_rrdfile_in = mdtm_rrdfile_open(filepath);
  mdtm_rrdfile_setformat(g_rrdfile_in, "nsid,size,xfr");
  g_rrddata_in = mdtm_rrddata_alloc(g_rrdfile_in);

  strcpy(fp, LOG_AGGR_RRD_OUT);
  g_rrdfile_out = mdtm_rrdfile_open(filepath);
  mdtm_rrdfile_setformat(g_rrdfile_out, "nsid,size,xfr");
  g_rrddata_out = mdtm_rrddata_alloc(g_rrdfile_out);

  strcpy(fp, LOG_AGGR_CSV_SNAP);
  g_csvfile = mdtm_csvfile_open(filepath);
  mdtm_csvfile_setformat(g_csvfile, "ts,sid,uid,t0,t1,src,dst,size,xfr,rate,dir,rrd_loc");

  strcpy(fp, LOG_AGGR_CSV_COMPL);
  g_csvfile_compl = mdtm_csvfile_open(filepath);
  mdtm_csvfile_setformat(g_csvfile_compl, "ts,sid,uid,t0,t1,src,dst,size,xfr,rate,dir,rrd_loc");

  strcpy(fp, LOG_CSV_SERVICE);
  g_csvfile_service = mdtm_csvfile_open(filepath);
  mdtm_csvfile_setformat(g_csvfile_service, "ThreadID,Type,Numa,Core,Device");
}

void
mdtm_monitor_log_flush(void *userdata)
{
  mdtm_csvfile_flush_data(g_csvfile);
  mdtm_csvfile_flush_data(g_csvfile_service);
  mdtm_rrdfile_write(g_rrdfile_in, g_rrddata_in);
  mdtm_rrdfile_write(g_rrdfile_out, g_rrddata_out);
}

void
mdtm_monitor_log_progress(void *user_data)
{
  mdtm_xmit_progress_t *xmit_progress = (mdtm_xmit_progress_t *)user_data;
  void *rrdfile, *data;

  // update individual rrd file
  data = xmit_progress->data.rrd;
  rrdfile = xmit_progress->file.rrd_indv;
  mdtm_rrddata_set(data, "sid", xmit_progress->sid);
  mdtm_rrddata_set(data, "uid", xmit_progress->uid);
  mdtm_rrddata_set(data, "src", xmit_progress->src);
  mdtm_rrddata_set(data, "dst", xmit_progress->dst);
  mdtm_rrddata_set(data, "size", xmit_progress->size);
  mdtm_rrddata_set(data, "xfr", xmit_progress->xfr);
  mdtm_rrdfile_write(rrdfile, data);
  //  mdtm_rrdfile_close(rrdfile);

  //update global rrd and csv files
  mdtm_monitor_log_update(xmit_progress);
}

void
mdtm_monitor_log_progress_finished(void *user_data)
{
  mdtm_xmit_progress_t *xmit_progress = (mdtm_xmit_progress_t *)user_data;
  void *rrdfile, *data;

  // update individual rrd file
  data = xmit_progress->data.rrd;
  rrdfile = xmit_progress->file.rrd_indv;
  mdtm_rrddata_set(data, "sid", xmit_progress->sid);
  mdtm_rrddata_set(data, "uid", xmit_progress->uid);
  mdtm_rrddata_set(data, "src", xmit_progress->src);
  mdtm_rrddata_set(data, "dst", xmit_progress->dst);
  mdtm_rrddata_set(data, "size", xmit_progress->size);
  mdtm_rrddata_set(data, "xfr", xmit_progress->size);
  mdtm_rrdfile_write(rrdfile, data);
  mdtm_rrdfile_close(rrdfile);

  //update global rrd and csv files
  mdtm_monitor_log_finish(xmit_progress);
  free(xmit_progress->data.rrd);
  free(xmit_progress->callback_handle);
  free(xmit_progress);
}
